#!/usr/bin/env python

from grp import getgrnam
import json
import os.path
from pwd import getpwnam
import subprocess
from textwrap import dedent


CLIENT_RELATION_TYPES = frozenset(['db', 'db-admin'])


def relation_ids(relation_types):
    relids = []
    for reltype in relation_types:
        relid_cmd_line = ['relation-ids', '--format=json', reltype]
        json_relids = subprocess.check_output(relid_cmd_line).strip()
        if json_relids:
            relids.extend(json.loads(json_relids))
    return relids


def relation_list(relation_id):
    """Return the list of units participating in the relation."""
    cmd = ['relation-list', '--format=json', '-r', relation_id]
    json_units = subprocess.check_output(cmd).strip()
    if json_units:
        return json.loads(json_units)
    return []


def relation_get(relation_id, unit):
    cmd = ['relation-get', '--format=json', '-r', relation_id, '-', unit]
    json_relation = subprocess.check_output(cmd)
    if json_relation:
        return json.loads(json_relation)
    return {}


def relation_set(relation_id, keyvalues):
    command = ['relation-set', '-r', relation_id]
    command.extend(
        ["{}='{}'".format(k, v or '') for k, v in keyvalues.items()])
    subprocess.check_call(command)


def all_relations(relation_types=CLIENT_RELATION_TYPES):
    for reltype in relation_types:
        for relid in relation_ids(relation_types=[reltype]):
            for unit in relation_list(relation_id=relid):
                yield reltype, relid, unit, relation_get(relid, unit)


def rebuild_all_relations():
    for _, relid, unit, relation in all_relations():
        juju_log(MSG_INFO, "{} {} {!r}".format(relid, unit, relation))
        if 'database' in relation:
            rebuild_relation(relid, unit, relation)


def rebuild_relation(relid, unit, relation):
    relname = relid.split(':')[0]
    unitname = unit.replace('/', '-')

    script_name = 'psql-{}-{}'.format(relname, unitname)
    build_script(script_name, relation)
    state = relation.get('state', None)
    if state and state != 'standalone':
        script_name = 'psql-{}-{}'.format(relname, state)
        build_script(script_name, relation)


def build_script(script_name, relation):
    # Install a wrapper to psql that connects it to the desired database
    # by default. One wrapper per unit per relation.
    bin_dir = os.path.expanduser('~ubuntu/bin')
    install_dir(bin_dir, owner="ubuntu", group="ubuntu", mode=0o700)
    script_path = os.path.join(bin_dir, script_name)
    pgpass_path = os.path.abspath('pgpass.{}'.format(script_name))
    script = dedent(r"""\
        #!/bin/sh
        exec env \
            PGHOST={host} PGPORT={port} PGDATABASE={database} \
            PGUSER={user} PGPASSFILE={pgpass} \
            psql $*
        """).format(
            host=relation['host'],
            port=relation['port'],
            database=relation['database'],
            user=relation['user'],
            pgpass=pgpass_path)
    juju_log(MSG_INFO, "Creating {}".format(script_path))
    install_file(
        script, script_path, owner="ubuntu", group="ubuntu", mode=0o700)

    # The wrapper requires access to the password, stored in a .pgpass
    # file so it isn't exposed in an environment variable or on the
    # command line.
    pgpass = "*:*:*:{user}:{password}".format(
        user=relation['user'], password=relation['password'])
    install_file(
        pgpass, pgpass_path, owner="ubuntu", group="ubuntu", mode=0o400)


def install_file(contents, dest, owner="root", group="root", mode=0o600):
    uid = getpwnam(owner)[2]
    gid = getgrnam(group)[2]
    dest_fd = os.open(dest, os.O_WRONLY | os.O_TRUNC | os.O_CREAT, mode)
    os.fchown(dest_fd, uid, gid)
    with os.fdopen(dest_fd, 'w') as destfile:
        destfile.write(str(contents))


def install_dir(dirname, owner="root", group="root", mode=0o700):
    command = [
        '/usr/bin/install',
        '-o', owner, '-g', group, '-m', oct(mode), '-d', dirname]
    subprocess.check_call(command)


MSG_CRITICAL = "CRITICAL"
MSG_DEBUG = "DEBUG"
MSG_INFO = "INFO"
MSG_ERROR = "ERROR"
MSG_WARNING = "WARNING"


def juju_log(level, msg):
    subprocess.call(['juju-log', '-l', level, msg])


if __name__ == '__main__':
    raise SystemExit(rebuild_all_relations())
